%% WaveAtomC_Complex matlab example 1
% In this example we do the following:
% 1) Generate some complex valued data
% 2) Save generated data into a file in the correct format.
%    Remember to check whehter to use intType = 'int64' or 'int32'!
% 3) Calculate forward WaveAtom transform.
% 4) Load calculated WA coefficients into Matlab and perform compression.
% 5) Write compressed coefficients into a file in the correct format.
% 6) Calculate inverse WaveAtom transform.
% 7) Load inverse WaveAtom result into Matlab.
% 8) Validate result

%% Generate some random complex valued data to use as input
%Number of elements in each dimension must be a power of 2
data = randn(8,1)+randn(8,1)*1i;

%% Define used partition type
partition = 1; % 1 = multiscale 2 = monoscale

%% Define filenames
inputdatafile = '../validation/data.dat';
coefficientfile = '../validation/coef.dat';
modifiedcoefficientfile = '../validation/modcoef.dat';
outputdatafile = '../validation/out.dat';

%% Define program locations
WA_fwa = '../WA_fwa'; % Forward WA transform
WA_iwa = '../WA_iwa'; % Inverse WA transform

%% Define datatypes used in input data
% intType should be the type corresponding to size_t c-type.
% Usually 'int32' for 32-bit system and 'int64' for 64-bit system.
% Currently WaveAtom package supports only dataType 'double' 
% It is straightforward to modify WaveAtom package code to handle 'float'
intType = 'int64';
dataType = 'double';

%% Define parameters concerning the use of FFTW
% See '../WA_testrun.sh' for explanations
savewisdom = 1;
readwisdom = 1;
fwawisdomfile = '../validation/fwawisdom.dat';
iwawisdomfile = '../validation/iwawisdom.dat';
fftwmeasure = 0;
fftwthreads = 1;
fftwtimelimit = -1;

%% Write data into a file
fprintf(1,'Writing input data into file %s.\n',inputdatafile)
tic();
WA_writeDatainput(inputdatafile,data,intType,dataType);
toc()

%% Calculate forward WA transform
fprintf(1,'Calculating forward WA transform.\n')
tic();
[fs,fw] = unix([WA_fwa,...
    ' -inputdatafile ', inputdatafile, ...
    ' -coefdatafile ', coefficientfile, ...
    ' -partition ', num2str(partition), ...
    ' -savewisdom ', num2str(savewisdom), ...
    ' -readwisdom ', num2str(readwisdom), ...
    ' -wisdomfile ', fwawisdomfile, ...
    ' -fftwmeasure ', num2str(fftwmeasure), ...
    ' -fftwthreads ', num2str(fftwthreads), ...
    ' -fftwtimelimit ', num2str(fftwtimelimit), ...
    ]);

if(fs~=0)
    fw
    error('Forward Wave Atom transform failed.')
end
toc()

%% Read calculated WA coefficients
fprintf(1,'Reading WA coefficients.\n')
tic();
[coefs, coefsDims] = WA_readCoefficients(coefficientfile,intType,dataType);
toc() 
 
%% Do compression
cutoff = 0.00; %Choose suitable cutoff value for coefficients (between 0 and 1).
if(cutoff > 0)
    fprintf('Compressing coefficients with cutoff %d\n',cutoff)
    fprintf(1,'Coefficient data L^2-norm: %d\n',norm(coefs))
    z = coefs;
    tmp = sort(abs(coefs(:)));
    tmp = tmp(end:-1:1);
    ene = tmp.^2;
    ene = cumsum(ene);
    loc = find(ene>(1-cutoff)*ene(end), 1 );
    val = tmp(loc);
    if (isempty(val))
        val = 0;
    end    
    z(abs(z)<val) = 0;
    fprintf(1, 'Compression rate %d\n', numel(z)/nnz(z));
    %% Write compressed coefficients into a file
    tic();
    WA_writeCoefficients(modifiedcoefficientfile,z,coefsDims,intType,dataType);
    toc()
else
    fprintf(1,'Coefficient data L^2-norm: %d\n',norm(coefs))
    modifiedcoefficientfile = coefficientfile;
end


%% Perform inverse WA transform
fprintf(1,'Calculating inverse WA transform.\n')
tic();
[is,iw] = unix([WA_iwa,...
    ' -coefdatafile ', modifiedcoefficientfile, ...
    ' -outputdatafile ', outputdatafile, ...
    ' -partition ', num2str(partition), ...
    ' -savewisdom ', num2str(savewisdom), ...
    ' -readwisdom ', num2str(readwisdom), ...
    ' -wisdomfile ', iwawisdomfile, ...
    ' -fftwmeasure ', num2str(fftwmeasure), ...
    ' -fftwthreads ', num2str(fftwthreads), ...
    ' -fftwtimelimit ', num2str(fftwtimelimit), ...
    ]);
if(is~=0)
    iw
    error('Backward Wave Atom transform failed.')
end
toc()

%% Read output data
fprintf(1,'Reading resulting output.\n')
tic();
output =  WA_readDatainput(outputdatafile,intType,dataType);
toc()

%% Validate
fprintf(1,'Validating result.\n')
tic();
error = data-output;

errorL2 = norm(error(:));
inputL2 = norm(data(:));
outputL2 = norm(output(:));

toc()
fprintf(1,'\nWave Transform error L^2-norm: %d\n', errorL2)
fprintf(1,'Input data L^2-norm: %d\n', inputL2)
fprintf(1,'Output data L^2-norm: %d\n', outputL2)
fprintf(1,'(Output-L^2 / Input-L^2)^2: %d\n',(outputL2/inputL2)^2)
fprintf(1,'(Error-L^2 / Input-L^2)^2: %d\n',(errorL2/inputL2)^2)

fprintf(1,'(Error-L^2 / Input-L^2)^2: should be %d\n',cutoff)
