#include "WA_compress.h"

int main(int argc, char* argv[])
{
  /*
    Read settings
   */

  char *coefdatafile = NULL;
  char *modcoefdatafile = NULL;
  double energyreduction = 0;

  if (getCharCmdArg(argc,argv,"coefdatafile",&coefdatafile) == 0){
    fprintf(stderr,"-coefdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"modcoefdatafile",&modcoefdatafile) == 0){
    fprintf(stderr,"-modcoefdatafile not specified.\n");
    return 1;
  }

  if (getDoubleCmdArg(argc,argv,"energyreduction",&energyreduction) == 0){
    fprintf(stderr,"-energyreduction not specified. (Must be between 0 and 1.)\n");
    return 1;
  }
  
  if(energyreduction < 0 || energyreduction>1){
    fprintf(stderr,"Given amount of energy to be reduced is not wetween 0 and 1. It is %f.\n",energyreduction);
    return 1;
  }

  /*
    Read data in, sort coefficients, and calculate cutoff coefficient.
    Release data and read data back in correct order and use cutoff coefficient to zero some of the coefficients.
   */

  inputData od;
  double complex *coef = readDoubleComplexCoefficientDatafile(coefdatafile,&od);
  if(coef==NULL){
    fprintf(stderr,"Failed to read coefficients.\n");
    return 1;
  }  
  printInputData(&od,0);      
  if(energyreduction > 0){    
    printf("Sorting %zu coefficients...\n",od.numberOfElements);
    qsort (coef, od.numberOfElements, sizeof(double complex), compare);
    printf("Sorting done.\n");  
    
    //Calculate energy
    double energy = 0;
    double temp;
    for(size_t x = 0; x<od.numberOfElements;x++){
      temp = cabs(coef[x]);
      coef[x] = temp;
      energy += temp*temp;
    }
    
    printf("Coefficient data L^2-norm: %f\n",sqrt(energy));
    
    double cutoff = energy * energyreduction;
    if(cutoff > 0){
      size_t pos = 0;
      while (cutoff > 0){
	temp = coef[pos++];
	cutoff -= temp*temp;
      }
      cutoff = coef[pos-1];
      
      printf("Cutoff coefficient: %f\n",cutoff);
      
      fftw_free(coef);
      coef = readDoubleComplexCoefficientDatafile(coefdatafile,&od);
      if(coef==NULL){
	fprintf(stderr,"Failed to read coefficients.\n");
	freeInputData(&od);
	return 1;
      }  
      
      size_t zeroedCoef = 0;
      for(size_t x = 0; x<od.numberOfElements;x++){
	if (cabs(coef[x]) < cutoff){
	  coef[x] = 0;
	  zeroedCoef++;
	}
      }
      printf("Compression rate: %f\n",((double)od.numberOfElements)/(od.numberOfElements-zeroedCoef));
      
    } else {
      fftw_free(coef);
      coef = readDoubleComplexCoefficientDatafile(coefdatafile,&od);
      if(coef==NULL){
	freeInputData(&od);    
	return 1;
      }  
    }
  }

  /*
    Write compressed coefficients into a file
   */
  
  FILE *fp = fopen(modcoefdatafile,"wb");
  if(fp==NULL){
    fprintf(stderr,"Could not open file %s to write modified coeffcients.\n", modcoefdatafile);
    fftw_free(coef); freeInputData(&od);    
    return 1;
  }

  if(!writeDimensionHeader(&od,fp)){
    fprintf(stderr,"Failed to write dimension information into file %s.\n",modcoefdatafile);
    freeInputData(&od);
    fclose(fp);
    return 1;
  }  
  
  if(fwrite(coef,sizeof(double complex),od.numberOfElements,fp)!=od.numberOfElements){
    fprintf(stderr,"Failed to write the modified coefficients.\n");
    fftw_free (coef); 
    freeInputData(&od);
    fclose(fp);          
    return 1;
  }
  
  fclose(fp);
  fftw_free (coef); 
  freeInputData(&od);

  return 0;
}

int compare (const void * a, const void * b)
{
  double dx = cabs(*(double complex *)a);
  double dy = cabs(*(double complex *)b);
  if (dx < dy) {
    return -1;
  } else if (dx > dy) {
    return +1;
  }
  return 0;
}

