#include "fwa.h"

int main(int argc, char* argv[])
{
  /*
    Read settings
   */

  char *inputdatafile = NULL;
  char *coefdatafile = NULL;
  int partition = 0;

  if (getCharCmdArg(argc,argv,"inputdatafile",&inputdatafile) == 0){
    fprintf(stderr,"-inputdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"coefdatafile",&coefdatafile) == 0){
    fprintf(stderr,"-coefdatafile not specified.\n");
    return 1;
  }

  if (getIntCmdArg(argc,argv,"partition",&partition) == 0){
    fprintf(stderr,"-partition not specified (1 = multiscale, 2 = monoscale).\n");
    return 1;
  }

  if(partition != 1 && partition !=2){
    fprintf(stderr,"Valid value for -partition is 1 (multiscale) or 2 (monoscale).\n");
    return 1;
  }

  bool uniformpartition = partition-1;    

  int savewisdom = 0;
  int readwisdom = 0;
  char *wisdomfile = NULL;

  getIntCmdArg(argc,argv,"savewisdom",&savewisdom);
  getIntCmdArg(argc,argv,"readwisdom",&readwisdom);

  if(savewisdom == 1 || readwisdom == 1){
    if(getCharCmdArg(argc,argv,"wisdomfile",&wisdomfile) == 0){
      fprintf(stderr,"-wisdomfile not specified even though -savewisdom or -readwisdom specified.\n");
      return 1;
    }
  }

  int fftwmeasure = 0;
  int fftwthreads = 1;
  double fftwtimelimit;

  getIntCmdArg(argc,argv,"fftwmeasure",&fftwmeasure);
  getIntCmdArg(argc,argv,"fftwthreads",&fftwthreads); 
  if(getDoubleCmdArg(argc,argv,"fftwtimelimit",&fftwtimelimit)==0){
    fftwtimelimit = FFTW_NO_TIMELIMIT;
  } else {
    printf("FFT planning time limit: %f\n",fftwtimelimit);
    fftw_set_timelimit(fftwtimelimit);
  }

  if (fftw_init_threads() == 0){
    fprintf(stderr,"Failed to initialize FFT threads.\n");
    return 1;
  }

  fftwthreads = fftwthreads < 1 ? 1 : fftwthreads;
  printf("Using max %d threads in FFT calculations.\n",fftwthreads);
  fftw_plan_with_nthreads(fftwthreads);
  
  unsigned fftw_flags;
  switch(fftwmeasure){
  case 1: fftw_flags = FFTW_MEASURE; break;
  case 2: fftw_flags = FFTW_PATIENT; break;
  case 3: fftw_flags = FFTW_EXHAUSTIVE; break;
  default: fftw_flags = FFTW_ESTIMATE; break;
  }

  /*
    Read input data and print input data dimensions
   */

  time_t t1,t2;
  clock_t c1,c2;
  double fwacputime = 0;
  int fwarealtime = 0;

  (void) time(&t1);
  c1 = clock();

  inputData id;    
  if(readDoubleComplexData(inputdatafile,&id)==false){
    fprintf(stderr,"Failed to read inputdata.\n");
    return 1;
  }
  printInputData(&id,0);  

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to read input data: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to read input data: %d seconds.\n",(int)(t2-t1));

  /*
    Read FFTW wisdom files
   */

  if(readwisdom == 1){
    printf("Reading FFT wisdom from file %s.\n", wisdomfile);
    if(fftw_import_wisdom_from_filename(wisdomfile)==0){
      fprintf(stderr,"Failed to import FFT wisdom. Continuing without wisdom.\n");
    }
  }
  
  /*
    Perform FFT to input data
   */      

  (void) time(&t1);
  c1 = clock();

  printf("Performing fft to input data.\n");
  if(fftND_data(&id,fftw_flags) == false){
    fprintf(stderr,"Failed to perform fft to input data.\n");
    freeInputData(&id);
    return 1;
  }
  printf("Done performing fft to input data.\n");

  c2 = clock();
  (void) time(&t2);
  fwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  fwarealtime += (int)(t2-t1);
  printf("Cpu time to perform FFT to input data: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to perform FFT to input data: %d seconds.\n",(int)(t2-t1));

  /*
    Calculate coefficients
   */

  bool (*fwa)(const inputData *, double complex**, bool,unsigned);

  if(id.numberOfDimensions == 1){
    fwa = fwa1;
  } else if (id.numberOfDimensions == 2){
    fwa = fwa2;
  } else if (id.numberOfDimensions == 3){
    fwa = fwa3;
  } else if (id.numberOfDimensions == 4){
    fwa = fwa4;
  } else if (id.numberOfDimensions == 5){
    fwa = fwa5;    
  } else {
    fprintf(stderr,"Currently supported dimensions are 1-5. Current dimension is %zu.",id.numberOfDimensions);
    freeInputData(&id);
    return 1;
  }  
  

  (void) time(&t1);
  c1 = clock();

  double complex *coefficients = NULL;
  if(fwa(&id,&coefficients,uniformpartition,fftw_flags)==false){
    fprintf(stderr,"Failed to calculate %zuD forward wave transform.\n",id.numberOfDimensions);
    freeInputData(&id);
    return 1;
  } 

  c2 = clock();
  (void) time(&t2);
  fwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  fwarealtime += (int)(t2-t1);

  printf("Cpu time to calculate forward transform: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to calculate forward transform: %d seconds.\n",(int)(t2-t1));


  printf("\nTime to calculate forward transform (FFT+fwa) (CPU): %3.1f ms\n",fwacputime);
  printf("Time to calculate forward transform (FFT+fwa) (Real): %d s\n\n",fwarealtime);

  /*
    Save calculated coefficients
   */
  
  (void) time(&t1);
  c1 = clock();

  FILE *cfout = fopen(coefdatafile,"wb");
  if(cfout == NULL){
    fprintf(stderr,"Failed to open outputfile %s.\n",coefdatafile);
    freeInputData(&id);    
    fftw_free(coefficients);
    return 1;
  }
      
  if(!writeDimensionHeader(&id,cfout)){
    fprintf(stderr,"Failed to write dimension information into file %s.\n",coefdatafile);
    fclose(cfout);
    freeInputData(&id);
    fftw_free(coefficients);
    return 1;
  }  

  if(fwrite(coefficients, sizeof(double complex),id.numberOfElements, cfout)!=id.numberOfElements){
    fprintf(stderr,"Failed to write the coefficients into file %s.\n",coefdatafile);
    fclose(cfout);
    freeInputData(&id);   
    fftw_free(coefficients);
    return 1;    
  } 
  fclose(cfout);  
  freeInputData(&id);    
  fftw_free(coefficients);

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to write coefficients to disk: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to write coefficients to disk: %d seconds.\n",(int)(t2-t1));


  /*
    Save FFTW wisdom
   */

  if(savewisdom == 1){
    printf("Saving FFT wisdom to file %s.\n", wisdomfile);
    if (fftw_export_wisdom_to_filename(wisdomfile) == 0){
      fprintf(stderr,"Failed to save FFT wisdom to file.\n");
      return 1;
    }
  }

  printf("%zuD Forward wave transform calculated succesfully!\n",id.numberOfDimensions);

  return 0;
}
