#include "iwa.h"

int main(int argc, char* argv[])
{
  /*
    Read settings
   */

  char *coefdatafile = NULL;
  char *outputdatafile = NULL;
  int partition = 0;

  if (getCharCmdArg(argc,argv,"coefdatafile",&coefdatafile) == 0){
    fprintf(stderr,"-coefdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"outputdatafile",&outputdatafile) == 0){
    fprintf(stderr,"-outputdatafile not specified.\n");
    return 1;
  }

  if (getIntCmdArg(argc,argv,"partition",&partition) == 0){
    fprintf(stderr,"-partition not specified (1 = multiscale, 2 = monoscale).\n");
    return 1;
  }

  if(partition != 1 && partition !=2){
    fprintf(stderr,"Valid value for -partition is 1 (multiscale) or 2 (monoscale).\n");
    return 1;
  }

  bool uniformpartition = partition-1;  

  int savewisdom = 0;
  int readwisdom = 0;
  char *wisdomfile = NULL;

  getIntCmdArg(argc,argv,"savewisdom",&savewisdom);
  getIntCmdArg(argc,argv,"readwisdom",&readwisdom);

  if(savewisdom == 1 || readwisdom == 1){
    if(getCharCmdArg(argc,argv,"wisdomfile",&wisdomfile) == 0){
      fprintf(stderr,"-wisdomfile not specified even though -savewisdom or -readwisdom specified.\n");
      return 1;
    }
  }

  int fftwmeasure = 0;
  int fftwthreads = 1;
  double fftwtimelimit;

  getIntCmdArg(argc,argv,"fftwmeasure",&fftwmeasure);
  getIntCmdArg(argc,argv,"fftwthreads",&fftwthreads); 
  if(getDoubleCmdArg(argc,argv,"fftwtimelimit",&fftwtimelimit)==0){
    fftwtimelimit = FFTW_NO_TIMELIMIT;
  } else {
    printf("FFT planning time limit: %f\n",fftwtimelimit);
    fftw_set_timelimit(fftwtimelimit);
  }

  if (fftw_init_threads() == 0){
    fprintf(stderr,"Failed to initialize FFT threads.\n");
    return 1;
  }

  fftwthreads = fftwthreads < 1 ? 1 : fftwthreads;
  printf("Using max %d threads in FFT calculations.\n",fftwthreads);
  fftw_plan_with_nthreads(fftwthreads);
  
  unsigned fftw_flags;
  switch(fftwmeasure){
  case 1: fftw_flags = FFTW_MEASURE; break;
  case 2: fftw_flags = FFTW_PATIENT; break;
  case 3: fftw_flags = FFTW_EXHAUSTIVE; break;
  default: fftw_flags = FFTW_ESTIMATE; break;
  }

  /*
    Read coefficients and print coefficients dimensions
   */

  time_t t1,t2;
  clock_t c1,c2;
  double iwacputime = 0;
  int iwarealtime = 0;

  (void) time(&t1);
  c1 = clock();

  inputData od;  
  double complex *coef = readDoubleComplexCoefficientDatafile(coefdatafile,&od);
  if(coef==NULL){
    fprintf(stderr,"Failed to read coefficients.\n");
    return 1;
  }  
  printInputData(&od,0);

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to read coefficient data: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to read coefficient data: %d seconds.\n",(int)(t2-t1));

  /*
    Read FFTW wisdom files
   */

  if(readwisdom == 1){
    printf("Reading FFT wisdom from file %s.\n", wisdomfile);
    if(fftw_import_wisdom_from_filename(wisdomfile)==0){
      fprintf(stderr,"Failed to import FFT wisdom. Continuing without wisdom.\n");
    }
  }

  /*
    Calculate inverse coefficients
   */

  bool (*iwa)(inputData*, double complex*, bool, unsigned);
  if(od.numberOfDimensions == 1){
    iwa = iwa1;
  } else if (od.numberOfDimensions == 2){
    iwa = iwa2;
  } else if (od.numberOfDimensions == 3){
    iwa = iwa3;
  } else if (od.numberOfDimensions == 4){
    iwa = iwa4;
  } else if (od.numberOfDimensions == 5){
    iwa = iwa5;
  } else {
    fprintf(stderr,"Currently supported dimensions are 1-5. Current dimension is %zu.",od.numberOfDimensions);
    fftw_free(coef);
    freeInputData(&od);
    return 1;
  }

  (void) time(&t1);
  c1 = clock();
  
  if(!iwa(&od,coef,uniformpartition,fftw_flags)){
    fprintf(stderr,"Failed to calculate inverse coefficients.\n");
    fftw_free(coef); 
    freeInputData(&od);
    return 1;
  }

  c2 = clock();
  (void) time(&t2);
  iwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  iwarealtime += (int)(t2-t1);

  printf("Cpu time to calculate inverse coefficients: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to calculate inverse coefficients: %d seconds.\n",(int)(t2-t1));

  fftw_free(coef); 

  /*
    Perform inverse FFT to inverse coefficients
   */      

  (void) time(&t1);
  c1 = clock();  

  printf("Performing inverse FFT to inverse coefficients.\n");
  if(ifftND_data(&od,fftw_flags) == false){
    fprintf(stderr,"Failed to perform inverse FFT to input data.\n");
    freeInputData(&od);
    return 1;
  }
  printf("Done performing inverse FFT to inverse coefficients.\n");

  c2 = clock();
  (void) time(&t2);
  iwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  iwarealtime += (int)(t2-t1);
  printf("Cpu time to perform inverse FFT to inverse coefficients: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to perform inverse FFT to inverse coefficients: %d seconds.\n",(int)(t2-t1));
  printf("\nTime to calculate inverse transform (FFT+iwa) (CPU): %3.1f ms\n",iwacputime);
  printf("Time to calculate inverse transform (FFT+iwa) (Real): %d s\n\n",iwarealtime);  

  /*
    Save calculated inverse transform
   */  

  if(!writeDoubleComplexDatafile(outputdatafile,&od)){
    fprintf(stderr,"Failed to write output to a file %s\n",outputdatafile);
    freeInputData(&od);
    return 1;      
  }

  freeInputData(&od);

  /*
    Save FFTW wisdom
   */

  if(savewisdom == 1){
    printf("Saving FFT wisdom to file %s.\n", wisdomfile);
    if (fftw_export_wisdom_to_filename(wisdomfile) == 0){
      fprintf(stderr,"Failed to save FFT wisdom to file.\n");
      return 1;
    }
  }

  printf("%zuD Inverse wave transform calculated succesfully!\n",od.numberOfDimensions);    
  
  
  return 0;
}
