#include "WA_validator.h"

int main(int argc, char* argv[])
{

  /*
    Read settings
   */

  char *inputdatafile = NULL;
  char *outputdatafile = NULL;
  int buffersize = 0;

  if (getCharCmdArg(argc,argv,"inputdatafile",&inputdatafile) == 0){
    fprintf(stderr,"-inputdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"outputdatafile",&outputdatafile) == 0){
    fprintf(stderr,"-outputdatafile not specified.\n");
    return 1;
  }

  if (getIntCmdArg(argc,argv,"buffersize",&buffersize) == 0){    
    buffersize = 2048;
    fprintf(stderr,"-buffersize not specified. Using -buffersize %d.\n",buffersize);
  }

  /*
    Read input and output data. Match dimension headers and after that calculate L^2 norms of input data, output data, and the error between those.
   */

  inputData id;
  inputData od;

  printf("Validating results...\n");

  initializeInputData(&id);
  initializeInputData(&od);

  FILE *idf = fopen(inputdatafile,"rb");
  if(idf==NULL){
    fprintf(stderr,"Failed to open file '%s'.\n",inputdatafile);
    return 1;
  }

  FILE *odf = fopen(outputdatafile,"rb");
  if(odf==NULL){
    fprintf(stderr,"Failed to open file '%s'.\n",outputdatafile);
    fclose(idf);
    return 1;
  }

  if (!readDimensionHeader(&id, idf) || !readDimensionHeader(&od, odf)){
    freeInputData(&id); freeInputData(&od);
    fclose(idf); fclose(odf);
    return 1;
  }

  id.numberOfElements = 1;
  for(size_t i=0;i<id.numberOfDimensions;i++){
    id.numberOfElements *= id.sizeOfDimension[i];    
  }
  
  od.numberOfElements = 1;
  for(size_t i=0;i<od.numberOfDimensions;i++){
    od.numberOfElements *= od.sizeOfDimension[i];    
  }

  printf("Size of input data:\n");
  printInputData(&id,0);
  printf("Size of output data:\n");
  printInputData(&od,0);

  if (id.numberOfDimensions != od.numberOfDimensions){
    printf("Input and output have different number of dimensions!\n");
    fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
    return 1;
  } else {
    for (size_t i = 0; i<id.numberOfDimensions; ++i){
      if(id.sizeOfDimension[i] != od.sizeOfDimension[i]){
	printf("Input and output have different size of dimension %zu!\n",i);
	fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
	return 1;
      }
    }
  }

  size_t VB = buffersize;

  double complex *iddata = (double complex*)malloc(VB*sizeof(double complex));
  double complex *oddata = (double complex*)malloc(VB*sizeof(double complex));

  if (iddata == NULL || oddata == NULL){
    fprintf(stderr,"Failed to allocate memory for %zu elements of wadatatype data (%zu).\n", 2*VB, 2*VB*sizeof(double complex));
    fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
    free(iddata);free(oddata);
    return 1;
  }  

  printf("Calculating L^2-norms...\n");
  double error = 0;  
  double inputdataL2 = 0;
  double outputdataL2 = 0;
  double temp;
  for (size_t i = 0; i<(size_t)(id.numberOfElements/VB); ++i){
    if(fread(iddata,sizeof(double complex),VB,idf)!=VB){
      fprintf(stderr,"Failed to read dataelements\n");
      fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
      free(iddata);free(oddata);
      return 1;
    }    

    if(fread(oddata,sizeof(double complex),VB,odf)!=VB){
      fprintf(stderr,"Failed to read dataelements\n");
      fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
      free(iddata);free(oddata);
      return 1;
    }

    for(size_t j = 0;j<VB;++j){      
      temp = cabs(iddata[j]-oddata[j]);
      error += temp*temp;
      temp = cabs(iddata[j]);
      inputdataL2 += temp*temp;

      temp = cabs(oddata[j]);
      outputdataL2 += temp*temp;
    }
  }
  size_t rest = id.numberOfElements % VB;

  if(fread(iddata,sizeof(double complex),rest,idf)!=rest){
    fprintf(stderr,"Failed to read dataelements\n");
    fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
    free(iddata);free(oddata);
    return 1;
  }
  
  if(fread(oddata,sizeof(double complex),rest,odf)!=rest){
    fprintf(stderr,"Failed to read dataelements\n");
    fclose(idf); fclose(odf); freeInputData(&id); freeInputData(&od);
    free(iddata);free(oddata);
    return 1;
  } 
  
  for(size_t j = 0;j<rest;++j){
    temp = cabs(iddata[j]-oddata[j]);    
    error += temp*temp;

    temp = cabs(iddata[j]);
    inputdataL2 += temp*temp;
    
    temp = cabs(oddata[j]);
    outputdataL2 += temp*temp;
  }  

  inputdataL2 = sqrt(inputdataL2);
  outputdataL2 = sqrt(outputdataL2);
  error = sqrt(error);

  fclose(idf);
  fclose(odf);

  freeInputData(&id);
  freeInputData(&od);

  printf("Wave Transform error L^2-norm: %f\n",error);
  printf("Input data L^2-norm: %f\n",inputdataL2);
  printf("Output data L^2-norm: %f\n",outputdataL2);
  printf("(Output-L^2 / Input-L^2)^2: %f\n",outputdataL2*outputdataL2/(inputdataL2*inputdataL2));
  printf("(Error-L^2 / Input-L^2)^2: %f\n\n",error*error/(inputdataL2*inputdataL2));

  return 0;
}
