#include "WA_readdata.h"

bool writeDoubleComplexDatafile(const char* dir, inputData *id)
{
  FILE *fp = fopen(dir,"wb");
  if(fp==NULL){
    fprintf(stderr,"Failed to open file %s for writing.\n",dir);
    return false;
  }  

  if(!writeDimensionHeader(id, fp)){
    fprintf(stderr,"Failed to write data dimensions.\n");
    return false;
  }
  
  if (fwrite(id->data,sizeof(double complex),id->numberOfElements,fp) != id->numberOfElements){
    fprintf(stderr,"Failed to write output data.\n");
    return false;
  }
  
  fclose(fp);
  return true;
  
}

double complex* readDoubleComplexCoefficientDatafile(const char* dir, inputData *id)
{
  initializeInputData(id);
  FILE *fp = fopen (dir , "rb" );
  if (fp==NULL) {
    fprintf(stderr,"Could not open file %s.\n",dir);
    return false;
  }

  if(!readDimensionHeader(id,fp)){
    fprintf(stderr,"Failed to read the dimension header of %s\n",dir);
    fclose(fp);
    return false;
  }
  
  id->numberOfElements = 1;
  
  for(size_t i=0;i<id->numberOfDimensions;i++){
    id->numberOfElements *= id->sizeOfDimension[i];    
  }

  CudaSafeCall(cudaHostAlloc((void**)&id->data,
			     id->numberOfElements*sizeof(double complex),
			     cudaHostAllocDefault)); 
    
  if (id->data == NULL){
    fprintf(stderr,"Failed to allocate memory for coefficients.\n");
    fclose(fp);
    freeInputData(id);    
    return NULL;
  }
  
  // allocate memory to contain the whole file: 
  //  double complex *data = (double complex*) malloc(id->numberOfElements*sizeof(double complex));

  double complex *data = NULL;
  CudaSafeCall(cudaHostAlloc((void**)&data,
			     id->numberOfElements*sizeof(double complex),
			     cudaHostAllocDefault));  


  if (data == NULL) {
    fprintf(stderr,"Could not allocate enough memory for data! Tried to allocate %zu bytes.\n",id->numberOfElements*sizeof(double complex)); 
    freeInputData(id);
    fclose (fp);
    return NULL;
  }

  // copy the file into the buffer:
  if (fread(data,sizeof(double complex),id->numberOfElements,fp) != id->numberOfElements) {
    fprintf(stderr,"Error wile reading data from file %s.\n",dir);
    CudaSafeCall(cudaFreeHost(data));
    freeInputData(id);
    fclose(fp);
    return NULL;
  }

  fclose (fp);

  return data;  
}

bool readDoubleComplexData(const char* dir, inputData *id)
{
  printf("Reading inputdatafile '%s'...\n",dir);
  
  initializeInputData(id);

  FILE *fp = fopen(dir,"rb");
  if(fp==NULL){
    fprintf(stderr,"Failed to open file '%s'.\n",dir);
    return false;
  }

  if(!readDimensionHeader(id,fp)){
    fprintf(stderr,"Failed to read the dimension header of %s\n",dir);
    fclose(fp);
    return false;
  }

  id->numberOfElements = 1;
  
  for(size_t i=0;i<id->numberOfDimensions;i++){
    id->numberOfElements *= id->sizeOfDimension[i];    
  }
  
  //Pinned memory
  /*
  CudaSafeCall(cudaHostAlloc((void**)&id->data,
			     id->numberOfElements*sizeof(double complex),
			     cudaHostAllocWriteCombined|
			     cudaHostAllocMapped));
  */

    
  CudaSafeCall(cudaHostAlloc((void**)&id->data,
			     id->numberOfElements*sizeof(double complex),
			     cudaHostAllocDefault)); 
  

  if (id->data == NULL){
    fprintf(stderr,"Failed to allocate memory for %zu elements of double complex.\n", id->numberOfElements);
    fclose(fp);
    freeInputData(id);    
    return false;
  }  
  
  if(fread(id->data,sizeof(double complex),id->numberOfElements,fp)!=id->numberOfElements){
    fprintf(stderr,"Failed to read dataelements\n");
    fclose(fp);
    freeInputData(id);
    return false;
  }

  fclose(fp);
  printf("Reading inputdatafile done.\n");
  
  return true;  
}


void printInputData(const inputData *id, size_t nShowElements)
{
  printf("Number of dimensions: %zu\n", id->numberOfDimensions);

  if(id->sizeOfDimension == NULL){
    return;
  }  

  for(size_t i = 0; i < id->numberOfDimensions; ++i){
    printf("\tDimension %zu: %zu\n", i,id->sizeOfDimension[i]);    
  }  

  printf("Number of physical elements: %zu\n",id->numberOfElements);

  if (nShowElements > id->numberOfElements) nShowElements = id->numberOfElements;

  for(size_t i = 0; i < nShowElements; ++i){
    double complex temp = id->data[i];
    printf("%f+%fI\n", creal(temp),cimag(temp));    
  }
  printf("\n");

  return;
}

void initializeInputData(inputData *id)
{
  id->numberOfDimensions = 0;
  id->numberOfElements = 0;
  id->sizeOfDimension = NULL;
  id->data = NULL;
}

void freeInputData(inputData *id)
{  
  if(id->sizeOfDimension != NULL){
    free(id->sizeOfDimension);
    id->sizeOfDimension = NULL;
  }

  if(id->data != NULL){
    CudaSafeCall(cudaFreeHost(id->data));
    id->data = NULL;
  }

  return;
}

bool readDimensionHeader(inputData *id, FILE *fp)
{
  if(fread(&id->numberOfDimensions,sizeof(size_t),1,fp)!=1){
    fprintf(stderr,"Failed to read the number of dimensions.\n");
    return false;
  }
  
  id->sizeOfDimension = (size_t*)malloc(id->numberOfDimensions*sizeof(size_t));
  if (id->sizeOfDimension == NULL){
    fprintf(stderr,"Failed to allocate memory for sizeOfDimensions.\n");
    return false;
  }
  
  if(fread(id->sizeOfDimension,sizeof(size_t),id->numberOfDimensions,fp) != id->numberOfDimensions){
    fprintf(stderr,"Failed to read the number of elements in dimensions.\n");
    freeInputData(id);
    return false;
  }
  return true;
}
  

bool writeDimensionHeader(const inputData *id, FILE *fp)
{
  
  if (fwrite(&(id->numberOfDimensions),sizeof(size_t),1,fp) != 1){
    return false;
  }

  if(fwrite(id->sizeOfDimension,sizeof(size_t),id->numberOfDimensions,fp)!=id->numberOfDimensions){
    return false;
  }

  return true;
}
