#include "WA_fwa.h"

bool fwa3(const inputData *id, double complex** coefficients, bool uniformpartition, cuDoubleComplex *d_inputdata)
{
  /*
    Check that input data has correct number of dimensions.
   */

  printf("Calculating 3D forward wave atom transform...\n");

  if(id->numberOfDimensions != 3){
    printf("Number of dimensions is not three! Instead, it is (%zu).\n", id->numberOfDimensions);
    return false;
  }

  /* 
     Reserve enough memory for different arrays, perform other initializations, and generate frequency partition.
   */

  double complex *coeff;
  CudaSafeCall(cudaHostAlloc((void**)&coeff,
			     id->numberOfElements*sizeof(double complex),
			     cudaHostAllocDefault)); 

  size_t sizeOfSmallestDimension = id->sizeOfDimension[0];
  for(size_t i = 1; i<3;++i){
    if (id->sizeOfDimension[i]<sizeOfSmallestDimension){
      sizeOfSmallestDimension=id->sizeOfDimension[i];
    }
  }

  size_t ratiosum = id->sizeOfDimension[0]/sizeOfSmallestDimension;
  size_t ratioprod = ratiosum; 
  for(size_t i = 1; i<3;++i){
    size_t ratio = id->sizeOfDimension[i]/sizeOfSmallestDimension;
    ratiosum += ratio;
    ratioprod *= ratio;
  }

  frequencyPartition freq_par;
  bool (*partitiofunc)(const size_t, frequencyPartition *);
  if (uniformpartition){
    partitiofunc = UniformFrequencyPartition;
  } else {
    partitiofunc = OrthogonalFrequencyPartition;
  }
	 
  if (!partitiofunc(sizeOfSmallestDimension, &freq_par)){
    cudaFreeHost(coeff);
    return false;
  }

  PrintFrequencyPartition(&freq_par);

  size_t maximumLengthOfM = freq_par.lengthOfM[0];
  size_t maximumElementsInCalculation = pow(freq_par.lengthOfM[freq_par.maxJ],3);
  if(freq_par.maxJ > 0 && pow(freq_par.lengthOfM[freq_par.maxJ-1],3)>maximumElementsInCalculation)
    maximumElementsInCalculation =  pow(freq_par.lengthOfM[freq_par.maxJ-1],3);

  for(size_t i = 1; i<=freq_par.maxJ;++i){
    if(maximumLengthOfM < freq_par.lengthOfM[i]){
      maximumLengthOfM = freq_par.lengthOfM[i];
    }      
  } 

  size_t mem_elems = ratiosum*maximumLengthOfM*(3<<freq_par.maxJ);
  size_t mem_size = sizeof(cuDoubleComplex)*mem_elems;

  cuDoubleComplex* d_mayercoef;
  cuDoubleComplex* d_mayercoefConj;
  //  cuDoubleComplex* d_inputdata;
  cuDoubleComplex* d_currentresult;
  int* d_wacoefStart;

  //Reserve device memory
  CudaSafeCall(cudaMalloc((void**)&d_mayercoef,mem_size));    
  CudaSafeCall(cudaMalloc((void**)&d_mayercoefConj,mem_size));    
  //  CudaSafeCall(cudaMalloc((void**)&d_inputdata,id->numberOfElements*sizeof(cuDoubleComplex)));    
  CudaSafeCall(cudaMalloc((void**)&d_wacoefStart,3*maximumLengthOfM*sizeof(int)));

  size_t freeMem;
  size_t totalMem;
  CudaSafeCall(cudaMemGetInfo(&freeMem,&totalMem));
  size_t maximumBlocksInOneGo = freeMem/((1<<((freq_par.maxJ+1)*3))*ratioprod*sizeof(cuDoubleComplex));  
  
  if(maximumBlocksInOneGo > maximumElementsInCalculation){
    maximumBlocksInOneGo = maximumElementsInCalculation;
  }

  //Try enabling this if you run out of memory when doing fft on device
  maximumBlocksInOneGo >>= 1;

  if(maximumBlocksInOneGo < 1){
    maximumBlocksInOneGo = 1;
  }


  CudaSafeCall(cudaMalloc((void**)&d_currentresult,maximumBlocksInOneGo*(1<<((freq_par.maxJ+1)*3))*ratioprod*sizeof(cuDoubleComplex)));

  /*
  printf("maximumBlocksInOneGo: %zu\n",maximumBlocksInOneGo);
  printf("mem for cur result: %zu bytes\n",(1<<((freq_par.maxJ+1)*3))*ratioprod*sizeof(cuDoubleComplex));	       
  */
	       
  double complex* wacoef = (double complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(double complex));  
  double complex* wacoefConj = (double complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(double complex));  
  int* wacoefStart = (int*) malloc(3*maximumLengthOfM*sizeof(int));  
  int *sijoituskohtaModJ = (int*)malloc(3*sizeof(int));
  int *sijoituskohtaModDim = (int*)malloc(3*sizeof(int));

  int *nFftwf = malloc(3*sizeof(int));

  size_t *ratioNcoef = malloc(3*sizeof(size_t));
 
  if (wacoef == NULL || wacoefConj == NULL || wacoefStart == NULL || 
      sijoituskohtaModJ == NULL || sijoituskohtaModDim == NULL ||
      nFftwf == NULL || ratioNcoef == NULL){
    printf("Error when reserving memory.");
    FreeFrequencyPartition(&freq_par);
    CudaSafeCall(cudaFree(d_mayercoef));
    CudaSafeCall(cudaFree(d_mayercoefConj));
    //    CudaSafeCall(cudaFree(d_inputdata));
    CudaSafeCall(cudaFree(d_currentresult));
    CudaSafeCall(cudaFree(d_wacoefStart));
    free(wacoef);
    free(wacoefConj);
    free(wacoefStart);
    free(sijoituskohtaModJ);
    free(sijoituskohtaModDim);
    free(nFftwf);
    free(ratioNcoef);
    cudaFreeHost(coeff);
    return false;
  }    



  /*
  //Copu input data to device
  CudaSafeCall(cudaMemcpy(d_inputdata, id->data, id->numberOfElements*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));  
  */

  size_t pos;
  int posStart;
  
  for(size_t h = 0; h<3;++h){
    ratioNcoef[h] = id->sizeOfDimension[h]/sizeOfSmallestDimension;
  }

  double complex* tempmemorystart = (double complex*)coeff;

  /*
    Loop over scale j
        Calculate required Mayer's Wavelet values for scale j.
	Loop over right/left bumps i=(i1,i2,i3)
	    Calculate coefficient set for fixed j and i by
	    looping over frequency partition m=(m1,m2,m3) and by
	    looping over 3D frequency area specified by (j,m) using index s=(s1,s2,s3).
	    Here m specifies CUDA grid size and blocks go over s.
	End
	Perform FFT to each calculated coefficient set j and m.
	Normalize each coefficient set to get final coefficients corresponding to j and m.
    End
    Return calculated coefficents.
   */
	       
  for (size_t j=0; j<=freq_par.maxJ;++j){

    printf("Scale j=%zu...\n",j);

    size_t lm = freq_par.lengthOfM[j];
    size_t sm = freq_par.startOfM[j];
    
    for(size_t h = 0; h<3;++h){
      ratioNcoef[h] <<=1;
    }      

    if(lm==0) continue;

    size_t nelemsj = (1<<((j+1)*3))*ratioprod;
    double b = 1<<j;

    for(size_t i = 0; i<3; ++i){
      nFftwf[i] =  (int)ratioNcoef[i];
    }   

    pos = 0;
    posStart = 0;

    for (size_t curDim = 0; curDim <3;++curDim){	
      bool wacoefsCalculatedAlready = false;            
      size_t copypos = 0;
      for (size_t d = 0; d<curDim;++d){
	if(id->sizeOfDimension[d] == id->sizeOfDimension[curDim]){
	  wacoefsCalculatedAlready = true;
	  memcpy(&wacoefStart[posStart],&wacoefStart[d*lm],sizeof(int)*lm);
	  memcpy(&wacoef[pos],&wacoef[copypos],sizeof(double complex)*ratioNcoef[d]*lm);
	  memcpy(&wacoefConj[pos],&wacoefConj[copypos],sizeof(double complex)*ratioNcoef[d]*lm);
	  pos += ratioNcoef[d]*lm;
	  posStart += lm;
	  break;
	}
	copypos += ratioNcoef[d]*lm;
      }
     
      if(wacoefsCalculatedAlready == false){      
	double d,t;
	double complex c;
	for (size_t m = 0; m<freq_par.lengthOfM[j]/2;m++){
	  double fm = id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  double dx = 1./(fm*b);

	  fm *= (2*m-2/3.)*b;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m;
	  t = M_PI_2*(d-0.5);

	  size_t tpos = pos;
	  double complex temp;

	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if(4/3. < d || d < -2/3.) temp = 0;
	    else if (2/3. < d) temp=c*sin(M_PI_4*(1+cos(M_PI*(1-1.5*d))));
	    else temp= c*cos(M_PI_4*(1-cos(M_PI*(0.5-0.75*d))));
	    wacoef[ratioNcoef[curDim]-1+tpos-i] = temp;
	    wacoefConj[pos++] = conj(temp);
	    d += dx;  
	    t += M_PI_2*dx;
	  }	

	  fm += 4/3.*b*id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m-1;
	  t = M_PI_2*(d-0.5);
	  tpos = pos;
	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if (d < -1/3. || 5/3. < d) temp = 0;
	    if (d < 1/3.) temp = c*sin(M_PI_4*(1+cos(M_PI*(1.5*d-0.5))));
	    else temp = c*cos(M_PI_4*(1-cos(M_PI*(0.75*d-0.25))));
	    wacoef[ratioNcoef[curDim]-1+tpos-i] = temp;
	    wacoefConj[pos++] = conj(temp);
	    d += dx;  
	    t += M_PI_2*dx;
	  }		
	}
      }
    }

    //Copy host data to device

    CudaSafeCall(cudaMemcpy(d_wacoefStart, wacoefStart, 3*maximumLengthOfM*sizeof(int), cudaMemcpyHostToDevice));    
    CudaSafeCall(cudaMemcpy(d_mayercoef, wacoef, ratiosum*lm*(2<<j)*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
    CudaSafeCall(cudaMemcpy(d_mayercoefConj, wacoefConj, ratiosum*lm*(2<<j)*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));

    size_t blocksLeft = lm*lm*lm;
    size_t actuallyCalcBlocksAlready = 0;
    while(blocksLeft>0){
      size_t startBlock = lm*lm*lm-blocksLeft;
      size_t numberOfBlocks = blocksLeft < maximumBlocksInOneGo ? blocksLeft : maximumBlocksInOneGo;

      size_t actuallyCalcBlocks=0;
      
      size_t m3 = (startBlock+numberOfBlocks)%lm;
      size_t m2 = ((startBlock+numberOfBlocks-m3)/lm)%lm;
      size_t m1 = (startBlock+numberOfBlocks-m3-m2*lm)/(lm*lm);

      if(m1>0 || m2>0 || m3>0)      
      if (m1 < sm && m2 < sm && m3 < sm){
	if (m2 > 0){
	  m2 -= 1;
	} else if(m1 > 0){
	  m2 = lm-1;
	  m1 -= 1;
	}
	m3=lm;
      }

      if(m1>=sm || m2>=sm || m3>=sm){	
	if(m1<sm){
	  actuallyCalcBlocks += (lm*lm-sm*sm)*m1;
	  if(m2<sm){
	    actuallyCalcBlocks += (m2*(lm-sm)+m3-sm); 
	  } else {
	    actuallyCalcBlocks += ((lm-sm)*sm + (m2-sm)*lm+m3);
	  }
	} else {
	  actuallyCalcBlocks += ((lm*lm-sm*sm)*sm + ((m1-sm)*lm+m2)*lm+m3);
	}
      }

      m3 = (startBlock)%lm;
      m2 = ((startBlock-m3)/lm)%lm;
      m1 = (startBlock-m3-m2*lm)/(lm*lm);

      if(m1>0 || m2>0 || m3>0)
      if (m1 < sm && m2<sm && m3<sm){
	if (m2>0){
	  m2 -= 1;
	} else if(m1>0){
	  m2 = lm-1;
	  m1 -= 1;
	}
	m3=lm;
      }

      if(m1>=sm || m2>=sm || m3>=sm){	
	if(m1<sm){
	  actuallyCalcBlocks -= (lm*lm-sm*sm)*m1;
	  if(m2<sm){
	    actuallyCalcBlocks -= (m2*(lm-sm)+m3-sm); 
	  } else {
	    actuallyCalcBlocks -= ((lm-sm)*sm + (m2-sm)*lm+m3);
	  }
	} else {
	  actuallyCalcBlocks -= ((lm*lm-sm*sm)*sm + ((m1-sm)*lm+m2)*lm+m3);
	}
      }

      CudaSafeCall(cudaMemset(d_currentresult, 0, actuallyCalcBlocks*nelemsj*sizeof(cuDoubleComplex)));

      cuDoubleComplex* d_dataws1 = d_mayercoefConj;		
      for(int i1 = 0; i1<2; ++i1){
	
	cuDoubleComplex* d_dataws2 = d_mayercoefConj;		
	for(int i2 = 0; i2<2; ++i2){	    
	  
	  cuDoubleComplex* d_dataws3 = d_mayercoefConj;		
	  for(int i3 = 0; i3<2; ++i3){	    
	    kernel3D_gpu_wrapper(i1,i2,i3,
				 sm,lm, 
				 startBlock, numberOfBlocks,
				 actuallyCalcBlocksAlready,
				 id->sizeOfDimension[0],
				 id->sizeOfDimension[1],
				 id->sizeOfDimension[2],
				 ratioNcoef[0],
				 ratioNcoef[1],
				 ratioNcoef[2],
				 nelemsj, 
				 d_inputdata,
				 d_currentresult,
				 d_dataws1, 
				 d_dataws2,
				 d_dataws3,
				 d_wacoefStart
				 );
	    CudaCheckError();	
	    
	    d_dataws3 = d_mayercoef;
	  }
	  d_dataws2 = d_mayercoef;
	}
	d_dataws1 = d_mayercoef;		
      }
      
      //Do FFT on card      
      cufftHandle plan;
      CufftSafeCall(cufftPlanMany(&plan, 3,nFftwf,nFftwf,1,nelemsj,nFftwf,1,nelemsj,CUFFT_Z2Z,actuallyCalcBlocks));
      CufftSafeCall(cufftExecZ2Z(plan, (cufftDoubleComplex *)d_currentresult, (cufftDoubleComplex *)d_currentresult, CUFFT_INVERSE));
      CufftSafeCall(cufftDestroy(plan));

      thrustMultiplyOnDevice(d_currentresult,nelemsj*actuallyCalcBlocks,1./sqrt(nelemsj));
            
      //Copy data from device	
      CudaSafeCall(cudaMemcpyAsync(tempmemorystart, d_currentresult, actuallyCalcBlocks*nelemsj*sizeof(cuDoubleComplex), cudaMemcpyDeviceToHost,0));      
      /*
	CudaSafeCall(cudaMemcpy(tempmemorystart, d_currentresult, actuallyCalcBlocks*nelemsj*sizeof(cuDoubleComplex), cudaMemcpyDeviceToHost));
      */

      blocksLeft -= numberOfBlocks;
      tempmemorystart += actuallyCalcBlocks*nelemsj;
      actuallyCalcBlocksAlready += actuallyCalcBlocks;

    }

  }
  CudaSafeCall(cudaFree(d_mayercoef));
  CudaSafeCall(cudaFree(d_mayercoefConj));
  //  CudaSafeCall(cudaFree(d_inputdata));
  CudaSafeCall(cudaFree(d_currentresult));
  CudaSafeCall(cudaFree(d_wacoefStart));
  free(wacoef);    
  free(wacoefConj);    
  free(wacoefStart);
  free(sijoituskohtaModJ);
  free(sijoituskohtaModDim);
  free(nFftwf);
  FreeFrequencyPartition(&freq_par);
  free(ratioNcoef);

  (*coefficients) = coeff;

  return true;
}
