#include "fwa.h"

int main(int argc, char* argv[])
{
  /*
    Read settings
   */

  char *inputdatafile = NULL;
  char *coefdatafile = NULL;
  int partition = 0;

  if (getCharCmdArg(argc,argv,"inputdatafile",&inputdatafile) == 0){
    fprintf(stderr,"-inputdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"coefdatafile",&coefdatafile) == 0){
    fprintf(stderr,"-coefdatafile not specified.\n");
    return 1;
  }

  if (getIntCmdArg(argc,argv,"partition",&partition) == 0){
    fprintf(stderr,"-partition not specified (1 = multiscale, 2 = monoscale).\n");
    return 1;
  }

  if(partition != 1 && partition !=2){
    fprintf(stderr,"Valid value for -partition is 1 (multiscale) or 2 (monoscale).\n");
    return 1;
  }

  bool uniformpartition = partition-1;  

  /*
    Initialize GPU
   */
  time_t t1,t2;
  clock_t c1,c2;
  cudaEvent_t start, stop;
  float elapsedTime;    
  
  double fwacputime = 0;
  double fwagputime = 0;
  int fwarealtime = 0;

  (void) time(&t1);
  c1 = clock();

  printf("Initializing GPU...\n");
  int devId = 0;
  if(getIntCmdArg(argc,argv,"device",&devId)){
    devId = gpuDeviceInit(devId);
  } else {
    devId = findCudaDevice();
  }
  if (devId < 0) {
    fprintf(stderr,"Failed to initialize a GPU.\n");
    return 1;
  }
  CudaSafeCall(cudaEventCreate(&start));
  CudaSafeCall(cudaEventCreate(&stop));    
  c2 = clock();
  (void) time(&t2);

  printf("GPU initialization time (CPU): %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("GPU initialization time (Real): %d seconds.\n",(int)(t2-t1));

  thrustVersion();


  /*
    Read input data and print input data dimensions
   */

  (void) time(&t1);
  c1 = clock();

  inputData id;    
  if(readDoubleComplexData(inputdatafile,&id)==false){
    fprintf(stderr,"Failed to read inputdata.\n");
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }
  printInputData(&id,0);  

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to read input data: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to read input data: %d seconds.\n",(int)(t2-t1));

  
  /*
    Perform FFT to input data
   */      

  (void) time(&t1);
  c1 = clock();
  CudaSafeCall(cudaEventRecord(start,0));

  printf("Performing fft to input data.\n");
  //Device memory for data
  size_t mem_size = sizeof(cuDoubleComplex)*id.numberOfElements;
  cuDoubleComplex* d_data;
  CudaSafeCall(cudaMalloc((void**)&d_data,mem_size));
  //Copy host data to device
  CudaSafeCall(cudaMemcpy(d_data, id.data, mem_size, cudaMemcpyHostToDevice));

  if(fft_CUDA_data(&id,d_data) == false){
    fprintf(stderr,"Failed to perform fft to input data.\n");
    freeInputData(&id);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }
  CudaSafeCall(cudaEventRecord(stop,0));
  CudaSafeCall(cudaEventSynchronize(stop));
  c2 = clock();
  (void) time(&t2);
  CudaSafeCall(cudaEventElapsedTime(&elapsedTime,start,stop));

  fwagputime += elapsedTime;
  fwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  fwarealtime += (int)(t2-t1);


  printf("Time to calculate FFT (GPU): %3.1f ms\n",elapsedTime);
  printf("Time to calculate FFT (CPU): %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Time to calculate FFT (Real): %d seconds.\n",(int)(t2-t1));
  printf("Done performing fft to input data.\n");

  /*
    Calculate coefficients
   */

  bool (*fwa)(const inputData *, double complex**, bool,cuDoubleComplex*);

  if (id.numberOfDimensions == 2){
    fwa = fwa2;
  } else if (id.numberOfDimensions == 3){
    fwa = fwa3;
  } else {
    fprintf(stderr,"Supported dimensions in GPU case are only 2 and 3. Current dimension is %zu.",id .numberOfDimensions);
    freeInputData(&id);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }


  (void) time(&t1);
  c1 = clock();
  CudaSafeCall(cudaEventRecord(start,0));

  double complex *coefficients = NULL;
  if(fwa(&id,&coefficients,uniformpartition,d_data)==false){
    fprintf(stderr,"Failed to calculate %zuD forward wave transform.\n",id.numberOfDimensions);
    freeInputData(&id);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  } 

  CudaSafeCall(cudaEventRecord(stop,0));
  CudaSafeCall(cudaEventSynchronize(stop));
  c2 = clock();
  (void) time(&t2);

  CudaSafeCall(cudaFree(d_data));

  CudaSafeCall(cudaEventElapsedTime(&elapsedTime,start,stop));
  fwagputime += elapsedTime;
  fwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  fwarealtime += (int)(t2-t1);
  printf("GPU time to calculate forward transform: %3.1f ms\n",elapsedTime);
  printf("Cpu time to calculate forward transform: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to calculate forward transform: %d seconds.\n",(int)(t2-t1));

  printf("\nTime to calculate forward transform (FFT+fwa) (GPU): %3.1f ms\n",fwagputime); 
  printf("Time to calculate forward transform (FFT+fwa) (CPU): %3.1f ms\n",fwacputime);
  printf("Time to calculate forward transform (FFT+fwa) (Real): %d s\n\n",fwarealtime);

  /*
    Save calculated coefficients
   */
 
  (void) time(&t1);
  c1 = clock();

  FILE *cfout = fopen(coefdatafile,"wb");
  if(cfout == NULL){
    fprintf(stderr,"Failed to open outputfile %s.\n",coefdatafile);
    freeInputData(&id);    
    cudaFreeHost(coefficients);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }
      
  if(!writeDimensionHeader(&id,cfout)){
    fprintf(stderr,"Failed to write dimension information into file %s.\n",coefdatafile);
    fclose(cfout);
    freeInputData(&id);
    cudaFreeHost(coefficients);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }  

  if(fwrite(coefficients, sizeof(double complex),id.numberOfElements, cfout)!=id.numberOfElements){
    fprintf(stderr,"Failed to write the coefficients into file %s.\n",coefdatafile);
    fclose(cfout);
    freeInputData(&id);
    cudaFreeHost(coefficients);
    CudaSafeCall(cudaDeviceReset());
    return 1;    
  } 
  fclose(cfout);  

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to write coefficients to disk: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to write coefficients to disk: %d seconds.\n",(int)(t2-t1));


  freeInputData(&id);      
  cudaFreeHost(coefficients);
  CudaSafeCall(cudaDeviceReset());

  printf("%zuD Forward wave transform calculated succesfully!\n",id.numberOfDimensions);

  return 0;
}
