#include "WA_iwa.h"

bool iwa3(const inputData *id, double complex* coefficients, bool uniformpartition)
{
  /*
    Check that input data has correct number of dimensions.
   */

  printf("Calculating 3D inverse wave atom transform...\n");

  size_t rank = id->numberOfDimensions;
  if(rank != 3){
    printf("Number of dimensions is not three! Instead, it is (%zu).\n", rank);
    return false;
  }

  /* 
     Reserve enough memory for different arrays, perform other initializations, and generate frequency partition.
   */

  size_t sizeOfSmallestDimension = id->sizeOfDimension[0];
  for(size_t i = 1; i<rank;++i){
    if (id->sizeOfDimension[i]<sizeOfSmallestDimension){
      sizeOfSmallestDimension=id->sizeOfDimension[i];
    }
  }

  size_t ratiosum = id->sizeOfDimension[0]/sizeOfSmallestDimension;
  size_t ratioprod = ratiosum; 
  for(size_t i = 1; i<rank;++i){
    size_t ratio = id->sizeOfDimension[i]/sizeOfSmallestDimension;
    ratiosum += ratio;
    ratioprod *= ratio;
  }

  frequencyPartition freq_par;
  bool (*partitiofunc)(const size_t, frequencyPartition *);
  if (uniformpartition){
    partitiofunc = UniformFrequencyPartition;
  } else {
    partitiofunc = OrthogonalFrequencyPartition;
  }
	 
  if (!partitiofunc(sizeOfSmallestDimension, &freq_par)){
    return false;
  }

  PrintFrequencyPartition(&freq_par);

  size_t maximumLengthOfM = freq_par.lengthOfM[0];
  for(size_t i = 1; i<=freq_par.maxJ;++i){
    if(maximumLengthOfM < freq_par.lengthOfM[i]){
      maximumLengthOfM = freq_par.lengthOfM[i];
    }      
  } 

  size_t maximumElementsInScale = (1<<((freq_par.maxJ+1)*rank))*ratioprod*(pow(freq_par.lengthOfM[freq_par.maxJ],rank)-pow(freq_par.startOfM[freq_par.maxJ],rank));
  if(freq_par.maxJ>0){
    size_t temp = (1<<((freq_par.maxJ)*rank))*ratioprod*(pow(freq_par.lengthOfM[freq_par.maxJ-1],rank)-pow(freq_par.startOfM[freq_par.maxJ-1],rank));
    if(maximumElementsInScale < temp){
      maximumElementsInScale = temp;
    }
  }

  //Calculate how much memory is required  
  size_t mem_elems = ratiosum*maximumLengthOfM*(2<<freq_par.maxJ);
  size_t mem_size = sizeof(cuDoubleComplex)*mem_elems;

  cuDoubleComplex* d_mayercoef;
  cuDoubleComplex* d_mayercoefConj;
  cuDoubleComplex* d_inputdata;
  cuDoubleComplex* d_outputdata;
  int* d_wacoefStart;

  //Reserve device memory
  CudaSafeCall(cudaMalloc((void**)&d_mayercoef,mem_size));    
  CudaSafeCall(cudaMalloc((void**)&d_mayercoefConj,mem_size));    
  CudaSafeCall(cudaMalloc((void**)&d_inputdata,maximumElementsInScale*sizeof(cuDoubleComplex)));    
  CudaSafeCall(cudaMalloc((void**)&d_wacoefStart,rank*maximumLengthOfM*sizeof(int)));
  CudaSafeCall(cudaMalloc((void**)&d_outputdata,id->numberOfElements*sizeof(cuDoubleComplex)));

  double complex* wacoef = (double complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(double complex));  
  double complex* wacoefConj = (double complex*) malloc(ratiosum*maximumLengthOfM*(2<<freq_par.maxJ)*sizeof(double complex));  
  int* wacoefStart = (int*) malloc(rank*maximumLengthOfM*sizeof(int));  
  int *sijoituskohtaModJ = (int*)malloc(rank*sizeof(int));
  int *sijoituskohtaModDim = (int*)malloc(rank*sizeof(int));

  int *nFftwf = malloc(rank*sizeof(int));

  size_t *ratioNcoef = malloc(rank*sizeof(size_t));
 
  if (wacoef == NULL || wacoefConj == NULL || wacoefStart == NULL || 
      sijoituskohtaModJ == NULL || sijoituskohtaModDim == NULL ||
      nFftwf == NULL || ratioNcoef == NULL){
    printf("Error when reserving memory.");
    FreeFrequencyPartition(&freq_par);
    CudaSafeCall(cudaFree(d_mayercoef));
    CudaSafeCall(cudaFree(d_mayercoefConj));
    CudaSafeCall(cudaFree(d_inputdata));
    CudaSafeCall(cudaFree(d_outputdata));
    CudaSafeCall(cudaFree(d_wacoefStart));
    free(wacoef);
    free(wacoefConj);
    free(wacoefStart);
    free(sijoituskohtaModJ);
    free(sijoituskohtaModDim);
    free(nFftwf);
    free(ratioNcoef);
    return false;
  }    

  CudaSafeCall(cudaMemset(d_outputdata, 0, id->numberOfElements*sizeof(cuDoubleComplex)));

  double complex* currentCoefficientPos = coefficients;
  size_t pos;
  int posStart;
  
  for(size_t h = 0; h<rank;++h){
    ratioNcoef[h] = id->sizeOfDimension[h]/sizeOfSmallestDimension;
  }

  /*
    Loop over scale j
        Calculate required Mayer's Wavelet values for scale j.
	Perform inverse FFT to each coefficient set (j, m).
	Normalize each coefficient set (j,m).
        Loop over right/left bumps i=(i1,i2,i3)
	    Calculate contribution to output for fixed j and i by
	    looping over frequency partition m=(m1,m2,m3) and by 
	    looping over 3D frequency area specified by (j,m) using index s=(s1,s2,s3).
	    Here m specifies CUDA grid size and blocks go over s.
	    We must be careful not to write into two same position in the CUDA code (the p loops take care of that).
	End
    End
    Return output.
   */

	       
  for (size_t j=0; j<=freq_par.maxJ;++j){
    printf("Scale j=%zu...\n",j);

    size_t lm = freq_par.lengthOfM[j];
    size_t sm = freq_par.startOfM[j];
    
    for(size_t h = 0; h<rank;++h){
      ratioNcoef[h] <<=1;
    }      

    if(lm==0) continue;
     size_t nelemsj = (1<<((j+1)*rank))*ratioprod;
    double b = 1<<j;

    for(size_t i = 0; i<rank; ++i){
      nFftwf[i] =  (int)ratioNcoef[i];
    }   

    pos = 0;
    posStart = 0;
    
    for (size_t curDim = 0; curDim <rank;++curDim){	
      bool wacoefsCalculatedAlready = false;            
      size_t copypos = 0;
      for (size_t d = 0; d<curDim;++d){
	if(id->sizeOfDimension[d] == id->sizeOfDimension[curDim]){
	  wacoefsCalculatedAlready = true;
	  memcpy(&wacoefStart[posStart],&wacoefStart[d*lm],sizeof(int)*lm);
	  memcpy(&wacoef[pos],&wacoef[copypos],sizeof(double complex)*ratioNcoef[d]*lm);
	  memcpy(&wacoefConj[pos],&wacoefConj[copypos],sizeof(double complex)*ratioNcoef[d]*lm);
	  pos += ratioNcoef[d]*lm;
	  posStart += lm;
	  break;
	}
	copypos += ratioNcoef[d]*lm;
      }
     
      if(wacoefsCalculatedAlready == false){      
	double d,t;
	double complex c;
	for (size_t m = 0; m<freq_par.lengthOfM[j]/2;m++){
	  double fm = id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  double dx = 1./(fm*b);

	  fm *= (2*m-2/3.)*b;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m;
	  t = M_PI_2*(d-0.5);

	  size_t tpos = pos;
	  double complex temp;

	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if(4/3. < d || d < -2/3.) temp = 0;
	    else if (2/3. < d) temp=c*sin(M_PI_4*(1+cos(M_PI*(1-1.5*d))));
	    else temp= c*cos(M_PI_4*(1-cos(M_PI*(0.5-0.75*d))));
	    wacoefConj[ratioNcoef[curDim]-1+tpos-i] = conj(temp);
	    wacoef[pos++] = temp;
	    d += dx;  
	    t += M_PI_2*dx;
	  }	

	  fm += 4/3.*b*id->sizeOfDimension[curDim]/sizeOfSmallestDimension;
	  wacoefStart[posStart] = ceil(fm);
	  d = wacoefStart[posStart++]*dx-2*m-1;
	  t = M_PI_2*(d-0.5);
	  tpos = pos;
	  for(size_t i = 0; i < ratioNcoef[curDim]; ++i){
	    c = cos(t)-sin(t)*I;
	    if (d < -1/3. || 5/3. < d) temp = 0;
	    if (d < 1/3.) temp = c*sin(M_PI_4*(1+cos(M_PI*(1.5*d-0.5))));
	    else temp = c*cos(M_PI_4*(1-cos(M_PI*(0.75*d-0.25))));
	    wacoefConj[ratioNcoef[curDim]-1+tpos-i] = conj(temp);
	    wacoef[pos++] = temp;
	    d += dx;  
	    t += M_PI_2*dx;
	  }		
	}
      }
    }

    //Copy host data to device
    CudaSafeCall(cudaMemcpy(d_wacoefStart, wacoefStart, rank*maximumLengthOfM*sizeof(int), cudaMemcpyHostToDevice));

    CudaSafeCall(cudaMemcpy(d_mayercoef, wacoef, ratiosum*lm*(2<<j)*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
    CudaSafeCall(cudaMemcpy(d_mayercoefConj, wacoefConj, ratiosum*lm*(2<<j)*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));

    int nValidMCombinations = pow(lm,rank)-pow(sm,rank);            
    //Copy required input data to device
    CudaSafeCall(cudaMemcpy(d_inputdata, currentCoefficientPos, nelemsj*nValidMCombinations*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
    
    //Do FFT on card            
    cufftHandle plan;
    CufftSafeCall(cufftPlanMany(&plan, rank,nFftwf,nFftwf,1,nelemsj,nFftwf,1,nelemsj,CUFFT_Z2Z,nValidMCombinations));
    CufftSafeCall(cufftExecZ2Z(plan, (cufftDoubleComplex *)d_inputdata, (cufftDoubleComplex *)d_inputdata, CUFFT_FORWARD));
    CufftSafeCall(cufftDestroy(plan));
    thrustMultiplyOnDevice(d_inputdata,nelemsj*nValidMCombinations,1./sqrt(nelemsj));
    //FFT on card ends
    
    currentCoefficientPos+=nelemsj*nValidMCombinations;   

    int actuallyCalcBlocksAlready = 0;
    size_t startBlock = 0;
    int halfgridsize = lm/2;
    size_t numberOfBlocks = halfgridsize*halfgridsize*halfgridsize;

    cuDoubleComplex* d_dataws1 = d_mayercoef;
    for(int i1 = 0; i1<2; ++i1){	
      cuDoubleComplex* d_dataws2 = d_mayercoef;		
      for(int i2 = 0; i2<2; ++i2){	    
	cuDoubleComplex* d_dataws3 = d_mayercoef;		
	for(int i3 = 0; i3<2; ++i3){	    	
	  for(int p1 = 0; p1<2; ++p1){
	    for(int p2 = 0; p2<2; ++p2){
	      for(int p3 = 0; p3<2; ++p3){
		inverse_kernel3D_gpu_wrapper(i1,i2,i3,	      
					     p1,p2,p3,
					     sm,lm, 
					     startBlock, numberOfBlocks,
					     actuallyCalcBlocksAlready,
					     id->sizeOfDimension[0],
					     id->sizeOfDimension[1],
					     id->sizeOfDimension[2],
					     ratioNcoef[0],
					     ratioNcoef[1],
					     ratioNcoef[2],
					     nelemsj, 
					     d_inputdata,
					     d_outputdata,
					     d_dataws1, 
					     d_dataws2,
					     d_dataws3,
					     d_wacoefStart
					     );
		CudaCheckError();	
	      }
	    }
	  }
	  d_dataws3 = d_mayercoefConj;
	}
	d_dataws2 = d_mayercoefConj;
      }
      d_dataws1 = d_mayercoefConj;	
    }        
  }

  //Copy data from device
  CudaSafeCall(cudaMemcpy(id->data, d_outputdata, id->numberOfElements*sizeof(cuDoubleComplex), cudaMemcpyDeviceToHost));    

  //Cleanup
  CudaSafeCall(cudaFree(d_mayercoef));
  CudaSafeCall(cudaFree(d_mayercoefConj));
  CudaSafeCall(cudaFree(d_inputdata));
  CudaSafeCall(cudaFree(d_outputdata));
  CudaSafeCall(cudaFree(d_wacoefStart));
  free(wacoef);    
  free(wacoefConj);    
  free(wacoefStart);
  free(sijoituskohtaModJ);
  free(sijoituskohtaModDim);
  free(nFftwf);
  FreeFrequencyPartition(&freq_par);
  free(ratioNcoef);

  return true;
}
