#include "iwa.h"

int main(int argc, char* argv[])
{
  /*
    Read settings
   */

  char *coefdatafile = NULL;
  char *outputdatafile = NULL;
  int partition = 0;

  if (getCharCmdArg(argc,argv,"coefdatafile",&coefdatafile) == 0){
    fprintf(stderr,"-coefdatafile not specified.\n");
    return 1;
  }

  if (getCharCmdArg(argc,argv,"outputdatafile",&outputdatafile) == 0){
    fprintf(stderr,"-outputdatafile not specified.\n");
    return 1;
  }

  if (getIntCmdArg(argc,argv,"partition",&partition) == 0){
    fprintf(stderr,"-partition not specified (1 = multiscale, 2 = monoscale).\n");
    return 1;
  }

  if(partition != 1 && partition !=2){
    fprintf(stderr,"Valid value for -partition is 1 (multiscale) or 2 (monoscale).\n");
    return 1;
  }

  bool uniformpartition = partition-1;  

  /*
    Initialize GPU
   */

  time_t t1,t2;
  clock_t c1,c2;
  cudaEvent_t start, stop;
  float elapsedTime;    
  
  double iwacputime = 0;
  double iwagputime = 0;
  int iwarealtime = 0;

  (void) time(&t1);
  c1 = clock();

  printf("Initializing GPU...\n");
  int devId = 0;
  if(getIntCmdArg(argc,argv,"device",&devId)){
    devId = gpuDeviceInit(devId);
  } else {
    devId = findCudaDevice();
  }
  if (devId < 0) {
    fprintf(stderr,"Failed to initialize a GPU.\n");
    return 1;
  }
  CudaSafeCall(cudaEventCreate(&start));
  CudaSafeCall(cudaEventCreate(&stop));    
  c2 = clock();
  (void) time(&t2);

  printf("GPU initialization time (CPU): %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("GPU initialization time (Real): %d seconds.\n",(int)(t2-t1));

  thrustVersion();


  /*
    Read coefficients and print coefficients dimensions
   */

  (void) time(&t1);
  c1 = clock();

  inputData od;
  double complex *coef = readDoubleComplexCoefficientDatafile(coefdatafile,&od);
  if(coef==NULL){
    fprintf(stderr,"Failed to read coefficients.\n");
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }  
  printInputData(&od,0);    

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to read coefficient data: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to read coefficient data: %d seconds.\n",(int)(t2-t1));

  /*
    Calculate inverse coefficients
   */

  bool (*iwa)(const inputData*, double complex*, bool);
  if (od.numberOfDimensions == 2){
    iwa = iwa2;
  } else if (od.numberOfDimensions == 3){
    iwa = iwa3; 
  } else {
    fprintf(stderr,"Supported dimensions in GPU case are only 2 and 3. Current dimension is %zu.",od.numberOfDimensions);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }

  (void) time(&t1);
  c1 = clock();
  CudaSafeCall(cudaEventRecord(start,0));  

  if(!iwa(&od,coef,uniformpartition)){
    fprintf(stderr,"Failed to calculate inverse transform.\n");
    freeInputData(&od);
    CudaSafeCall(cudaFreeHost(coef));
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }

  CudaSafeCall(cudaEventRecord(stop,0));
  CudaSafeCall(cudaEventSynchronize(stop));
  c2 = clock();
  (void) time(&t2);
  CudaSafeCall(cudaEventElapsedTime(&elapsedTime,start,stop));
  iwagputime += elapsedTime;
  iwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  iwarealtime += (int)(t2-t1);
  printf("GPU time to calculate inverse transform: %3.1f ms\n",elapsedTime);
  printf("Cpu time to calculate inverse transform: %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Real time to calculate inverse transform: %d seconds.\n",(int)(t2-t1));


  CudaSafeCall(cudaFreeHost(coef));

  /*
    Perform inverse FFT to inverse coefficients
   */      

  (void) time(&t1);
  c1 = clock();
  CudaSafeCall(cudaEventRecord(start,0));

  printf("Performing inverse fft to output data.\n");
  if(ifft_CUDA_data(&od) == false){
    fprintf(stderr,"Failed to perform inverse fft to input data.\n");
    freeInputData(&od);
    CudaSafeCall(cudaDeviceReset());
    return 1;
  }
  printf("Done performing inverse FFT to output data.\n");

  CudaSafeCall(cudaEventRecord(stop,0));
  CudaSafeCall(cudaEventSynchronize(stop));
  c2 = clock();
  (void) time(&t2);
  CudaSafeCall(cudaEventElapsedTime(&elapsedTime,start,stop));

  iwagputime += elapsedTime;
  iwacputime += ((double)(c2-c1)*1000)/CLOCKS_PER_SEC;
  iwarealtime += (int)(t2-t1);

  printf("Time to calculate inverse FFT (GPU): %3.1f ms\n",elapsedTime);
  printf("Time to calculate inverse FFT (CPU): %3.1f ms.\n",((double)(c2-c1)*1000)/CLOCKS_PER_SEC);
  printf("Time to calculate inverse FFT (Real): %d seconds.\n",(int)(t2-t1));

  printf("\nTime to calculate inverse transform (FFT+iwa) (GPU): %3.1f ms\n",iwagputime); 
  printf("Time to calculate inverse transform (FFT+iwa) (CPU): %3.1f ms\n",iwacputime);
  printf("Time to calculate inverse transform (FFT+iwa) (Real): %d s\n\n",iwarealtime);

  /*
    Save calculated inverse transform
   */  

  (void) time(&t1);
  c1 = clock();

  if(!writeDoubleComplexDatafile(outputdatafile,&od)){
    fprintf(stderr,"Failed to write output to a file %s\n",outputdatafile);
    freeInputData(&od);
    CudaSafeCall(cudaDeviceReset());
    return 1;      
  }
  freeInputData(&od);

  c2 = clock();
  (void) time(&t2);
  printf("Cpu time to write outputdata to disk: %3.1f ms.\n",((double)(c2-c1))*1000/CLOCKS_PER_SEC);
  printf("Real time to write outputdata to disk: %d seconds.\n",(int)(t2-t1));

  printf("%zuD Inverse wave transform calculated succesfully!\n",od.numberOfDimensions);    
  
  
  return 0;
}
