#!/bin/bash

#Set testrun parameters

#Specifies the number of elements in each of the dimensions while generating random data. The number of elements must be larger than 4 and a power of two. More dimensions can be added by just inserting space and the number of elements in the corresponding dimension. For example, "datadimensions='64 128'" corresponds to generating a 64x128 2D-array of random double complex numbers and "datadimensions='64 128 256'" corresponds to generating a 64x128x256 3D-array of random double complex numbers. WA_datagen adds also the correct dimension header to the generated data file.
datadimensions='64 128' 

#Specifies the used frequency partition. Allowed values are 1 and 2, where 1 corresponds to multiscale (fancy) and 2 corresponds to monoscale.
#If you are using multiscale the smallest dimension must be greater or equal to 4.
#If you are using monoscale the smallest dimension must be greater or equal to 8.
partition=1

#Specifies how much energy is reduced when compressing the coefficients (must be between 0 and 1).
compenergyreduction=0.01

#Specifies how many elements are read in while calculating L^2-norms in validation 
validationbuffersize=2048

#Following four parameters specify whether to use wisdom files in FFTW calculations or not. (1=YES, 0=NO)
fwareadwisdom=1
fwasavewisdom=1
iwareadwisdom=1
iwasavewisdom=1

#Choose which fftw_flag to use in when creating FFTW plan
#0=FFTW_ESTIMATE, 1=FFTW_MEASURE, 2=FFTW_PATIENT, 3=FFTW_EXHAUSTIVE
#If using something else than 0, you must use wisdom files and run fwa and iwa once before actually using them.
fftwmeasure=0

#Amount of time in seconds given to FFTW to generate plan. (-1 infinite)
fftwtimelimit=-1

#Number of threads used in FFTW calculations
fftwthreads=1

#File for the input data
inputdatafile=validation/data.dat

#File for the coefficients
coefdatafile=validation/coef.dat

#File for the compressed coefficients
modcoefdatafile=validation/modcoef.dat

#File for the output data
outputdatafile=validation/out.dat

#Files for the FFTW wisdom data
fwafftwisdomfile=validation/fwawisdom.dat
iwafftwisdomfile=validation/iwawisdom.dat

#Specify executables
WA_datagen=./WA_datagen
WA_fwa=./WA_fwa
WA_compress=./WA_compress
WA_iwa=./WA_iwa
WA_validator=./WA_validator

#Specify which executables to run. (1=RUN, 0=DO NOT RUN)
RUN_datagen=1
RUN_fwa=1
RUN_compress=1
RUN_iwa=1
RUN_validator=1

#Actual testrun
echo Running wave atom transform tester..
echo
echo Printing used parameters...
echo
echo Dimensions of data: $datadimensions
echo Partition: $partition
echo Energyreduction: $compenergyreduction
echo Databuffersize \in validation: $validationbuffersize
echo
echo Inputdatafile: $inputdatafile
echo Coefdatafile: $coefdatafile
echo Modcoefdatafile: $modcoefdatafile
echo Outputdatafile: $outputdatafile
echo FWA FFT wisdomfile: $fwafftwisdomfile
echo IWA FFT wisdomfile: $iwafftwisdomfile
echo
echo FWA use wisdom: $fwareadwisdom
echo FWA save wisdom: $fwasavewisdom
echo IWA use wisdom: $iwareadwisdom
echo IWA save wisdom: $iwasavewisdom
echo
echo FFT planning method: $fftwmeasure
echo FFT planning timelimit: $fftwtimelimit
echo FFT maxthreads: $fftwthreads
echo
echo WA_datagen: $WA_datagen
echo WA_fwa: $WA_fwa
echo WA_compress: $WA_compress
echo WA_iwa: $WA_iwa
echo WA_validator: $WA_validator
echo
echo RUN_datagen: $RUN_datagen
echo RUN_fwa: $RUN_fwa
echo RUN_compress: $RUN_compress
echo RUN_iwa: $RUN_iwa
echo RUN_validator: $RUN_validator
echo
echo Printing parameters done.
echo
echo Performing actual testing...

#Generate random input data
if test "$RUN_datagen" == "1" 
then
    echo Generating random data...
    echo $WA_datagen $inputdatafile $datadimensions

    time $WA_datagen $inputdatafile $datadimensions
    echo
fi

#Calculate forward Wave Atom transform
if test "$RUN_fwa" == "1"
then
    echo Calculating forward wave atom transform...
    echo $WA_fwa \
	-inputdatafile $inputdatafile \
        -coefdatafile  $coefdatafile \
        -partition     $partition \
        -savewisdom    $fwasavewisdom \
        -readwisdom    $fwareadwisdom \
	-wisdomfile    $fwafftwisdomfile \
        -fftwmeasure   $fftwmeasure \
        -fftwthreads   $fftwthreads \
        -fftwtimelimit $fftwtimelimit

    time $WA_fwa \
	-inputdatafile $inputdatafile \
        -coefdatafile  $coefdatafile \
        -partition     $partition \
        -savewisdom    $fwasavewisdom \
        -readwisdom    $fwareadwisdom \
	-wisdomfile    $fwafftwisdomfile \
        -fftwmeasure   $fftwmeasure \
        -fftwthreads   $fftwthreads \
        -fftwtimelimit $fftwtimelimit
    echo
fi

#Compress coefficients
if test "$RUN_compress" == "1"
then
    echo Compressing coefficients...
    echo $WA_compress \
	-coefdatafile    $coefdatafile \
        -modcoefdatafile $modcoefdatafile \
        -energyreduction $compenergyreduction

    time $WA_compress \
	-coefdatafile    $coefdatafile \
        -modcoefdatafile $modcoefdatafile \
        -energyreduction $compenergyreduction
    echo
fi

#Perform inverse Wave Atom transform to compressed coefficients
if test "$RUN_iwa" == "1" 
then
    if test "$RUN_compress" == "0"
    then	
	modcoefdatafile=$coefdatafile
    fi	
    echo Calculating inverse wave atom transform...
    echo $WA_iwa \
	-coefdatafile   $modcoefdatafile \
        -outputdatafile $outputdatafile \
        -partition      $partition \
        -savewisdom     $iwasavewisdom \
        -readwisdom     $iwareadwisdom \
	-wisdomfile     $iwafftwisdomfile \
        -fftwmeasure    $fftwmeasure \
        -fftwthreads    $fftwthreads \
        -fftwtimelimit  $fftwtimelimit    

    time $WA_iwa \
	-coefdatafile   $modcoefdatafile \
        -outputdatafile $outputdatafile \
        -partition      $partition \
        -savewisdom     $iwasavewisdom \
        -readwisdom     $iwareadwisdom \
	-wisdomfile     $iwafftwisdomfile \
        -fftwmeasure    $fftwmeasure \
        -fftwthreads    $fftwthreads \
        -fftwtimelimit  $fftwtimelimit    
    echo
fi

#Validate output against input
if test "$RUN_validator" == "1"
then
    echo Running validator...
    echo $WA_validator \
	-inputdatafile  $inputdatafile \
        -outputdatafile $outputdatafile \
        -buffersize     $validationbuffersize

    time $WA_validator \
	-inputdatafile  $inputdatafile \
        -outputdatafile $outputdatafile \
        -buffersize     $validationbuffersize
    echo
    if test "$RUN_compress" == "1"
    then
	echo \(Error-L^2 / Input-L^2\)^2 should be $compenergyreduction.
    else
	echo \(Error-L^2 / Input-L^2\)^2 should be 0.
    fi
fi

echo Wave atom tester finished.
echo

